/* ANTINAT
 * =======
 * This software is Copyright (c) 2003 Malcolm Smith.
 * No warranty is provided, including but not limited to
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * This code is licenced subject to the GNU General
 * Public Licence (GPL).  See the COPYING file for more.
 */

#include "../../../config.h"
#include <DSNetwork.h>
#include <DSConfigFile.h>
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include "../../Shared.h"
#include "../../Connection.h"
#include "../../loggers.h"
#include "../../moddecl.h"
#include "../../filters.h"

extern "C" {
int command_type()
{
	return 0x03;
}
}

BOOL forwardUDPData(Connection * tcpconn, SOCKET one, SOCKET two)
{
	char * tmpbuf;
	int tmplen;
	sl_t addrlen;
	BOOL finished;
	char szBuffer[8192];
	SOCKET biggest;
	int len;
	int diff;
	fd_set fds;
#ifdef WITH_IPV6
	struct sockaddr_in6 sa_local;
	struct sockaddr_in6 sa_remote;
#else
	SOCKADDR sa_local;
	SOCKADDR sa_remote;
#endif
	sl_t sl;
	/* Where did the most recent LAN packet come from? */
	SOCKADDR * remoteaddr;
	SOCKADDR * localaddr;
	SOCKET tcpsock;
	char * addrstring1;
	char * addrstring2;

	tcpsock = tcpconn->getSocket();

#ifdef WITH_DEBUG
	printf("Entering UDP forward\n");
#endif

	finished = FALSE;
	remoteaddr=NULL;
	localaddr=NULL;
	while (!finished)
	{
		barfedinbound:
		FD_ZERO (&fds);
		FD_SET (tcpsock, &fds);
		FD_SET (one, &fds);
		FD_SET (two, &fds);
		biggest=tcpsock;
		if (one > biggest)
			biggest = one;
		if (two > biggest)
			biggest = two;
		biggest++;
		/* Wait for message from either side */
		select (biggest, &fds, NULL, NULL, NULL);
		if (FD_ISSET (tcpsock, &fds))
		{
			/* We got something on TCP.  I read somewhere that
			 * some vendors have been extending the spec here.
			 * But it's not standard, and I'm ignoring it.  As
			 * soon as a read fails here, then the association is
			 * over.
			 */
			len=recv(tcpsock, szBuffer, sizeof(szBuffer) - 1, 0);
			if (len <= 0) {
				finished=TRUE;
#ifdef WITH_DEBUG
				printf("TCP close\n");
#endif
			} else {
#ifdef WITH_DEBUG
				printf("TCP activity\n");
#endif
			}
		}
		if (FD_ISSET (one, &fds))
		{
			sl=sizeof(sa_local);
#ifdef WITH_DEBUG
			printf("Outbound UDP packet\n");
#endif
			len = recvfrom (one, szBuffer,
				    sizeof(szBuffer) - 1, 0, (SOCKADDR *)&sa_local, &sl);
			if (len > 0)
			{

				unsigned char ch;
				proc_resolve resolver;
				if (!localaddr) localaddr=(SOCKADDR *)&sa_local;
				szBuffer[len] = '\0';
				Connection udpconn((SOCKADDR *)&sa_local, szBuffer, len);

				/* Check that packet came from same place as
				 * TCP connection */
				addrstring1=tcpconn->getSourceAddrString();
				addrstring2=udpconn.getSourceAddrString();
				diff=strcmp(addrstring1,addrstring2);
				free(addrstring1);
				free(addrstring2);
#ifdef WITH_DEBUG
				if (diff) printf("TCP source != UDP source\n");
#endif
				if (diff) goto barfedoutbound;

					
				if (!udpconn.IsAllowed(TYPE_SOURCE)) {
					goto barfedoutbound;
				}

				udpconn.setUser(tcpconn->getUser());

				/* Reserved */
				if (!udpconn.getChar(&ch)) goto barfedoutbound;
				if (ch!='\0') goto barfedoutbound;

				/* Reserved */
				if (!udpconn.getChar(&ch)) goto barfedoutbound;
				if (ch!='\0') goto barfedoutbound;

				/* Fragmentation (not handled yet) */
				if (!udpconn.getChar(&ch)) goto barfedoutbound;
				if (ch!='\0') goto barfedoutbound;

				/* Address type */
				if (!udpconn.getChar(&ch)) goto barfedoutbound;
				resolver=(proc_resolve)udpconn.getResolver(ch);
				if (!resolver) goto barfedoutbound;

				resolver(&udpconn);

				if (!tcpconn->getDestNullAddr()) {
					addrstring1=tcpconn->getDestAddrString();
					addrstring2=udpconn.getDestAddrString();
					diff=strcmp(addrstring1,addrstring2);
					free(addrstring1);
					free(addrstring2);
#ifdef WITH_DEBUG
					if (diff) printf("TCP dest != UDP dest\n");
#endif
					if (diff) goto barfedoutbound;
				}

				addrlen=sizeof(sa_remote);
				if (!udpconn.IsAllowed(TYPE_DEST)) {
					goto barfedoutbound;
				}
				if (remoteaddr) free(remoteaddr);
				udpconn.getDestSockaddr(&remoteaddr, &addrlen);
				tmpbuf=udpconn.getRemainingData();
				tmplen=udpconn.getRemainingLength();
				/* FIXME: FILTERING!! */
				sendto (two, tmpbuf, tmplen, 0, remoteaddr, addrlen);
#ifdef WITH_DEBUG
				printf("Send %i bytes local->remote\n",len);
#endif
			}
		}
		barfedoutbound:
		if (FD_ISSET (two, &fds))
		{
#ifdef WITH_DEBUG
			printf("inbound UDP packet\n");
#endif
			/* A message from the world, forward to the LAN */
			sl=sizeof(sa_remote);
			len = recvfrom (two, szBuffer,
				    sizeof(szBuffer) - 1, 0,
				    (SOCKADDR *)&sa_remote, &sl);
			if (len > 0)
			{
				char szBufferTwo[8300];
				char szTemp[16];
				proc_revresolve revresolve;
				if (!remoteaddr) goto barfedinbound;
				if (!localaddr) goto barfedinbound;
				/* Note: in this case, the source address is
				 * the recipient...*/
				Connection udpconn((SOCKADDR *)&sa_remote, szBufferTwo, sizeof(szBufferTwo));
				if (!tcpconn->getDestNullAddr()) {
					addrstring1=tcpconn->getDestAddrString();
					addrstring2=udpconn.getSourceAddrString();
					diff=strcmp(addrstring1,addrstring2);
					free(addrstring1);
					free(addrstring2);
					if (diff) goto barfedinbound;
				}
				szTemp[0]='\0'; /* Res */
				szTemp[1]='\0'; /* Res */
				szTemp[2]='\0'; /* Fragmentation */
				udpconn.sendData(szTemp, 3);
				revresolve=(proc_revresolve)udpconn.getRevResolver(localaddr->sa_family);
				if (!revresolve) goto barfedinbound;

				if (!revresolve(&udpconn, localaddr)) goto barfedinbound;
				udpconn.sendData(szBuffer, len);
				tmpbuf=udpconn.getCompleteData();
				tmplen=udpconn.getCompleteLength();
				addrlen=sizeof(sa_local);
				/* I'm assuming that there's no point filtering
				 * incoming UDP packets if an outgoing packet
				 * is allowed to reach the destination, the
				 * destination should be allowed to reach the
				 * source.*/
				sendto (one, tmpbuf, tmplen, 0, localaddr, addrlen);
#ifdef WITH_DEBUG
				printf("Send %i bytes remote->local\n",len);
#endif
			}
		}
	}
	if (remoteaddr) free(remoteaddr);
	return TRUE;
}

BOOL cpp_command(Connection * conn)
{
	SOCKET remotelisten, remotetalk;
#ifdef WITH_IPV6
	struct sockaddr_in6 sout;
#else
	SOCKADDR_IN sout;
#endif
	sl_t soutlen;
	SOCKADDR * addr;
	sl_t addrlen;

	if (!conn->getDestSockaddr(&addr, &addrlen)) return FALSE;
#ifdef WITH_DEBUG
	printf("Socket family: %x\n",addr->sa_family);
#endif
	remotelisten=socket(addr->sa_family, SOCK_DGRAM, IPPROTO_UDP);
#ifdef WITH_DEBUG
	{
		char * dest;
		dest=conn->getDestString();
		printf("Binding to %s\n",dest);
		free(dest);
	}
#endif
	/* FIXME: This hack listens on all interfaces on any port.
	 * Is this really what the spec had in mind? */
	memset(&sout, 0, sizeof(sout));
	((SOCKADDR *)&sout)->sa_family=addr->sa_family;
	soutlen=sizeof(sout);
	if (bind(remotelisten, (SOCKADDR *)&sout, soutlen)!=0) {
#ifdef WITH_DEBUG
		printf("Bind failed\n");
		perror("bind");
#endif
		CloseAndUnlink(remotelisten);
		conn->sendResponsev5(0x01);
		free(addr);
		return FALSE;
	}
	remotetalk=socket(addr->sa_family, SOCK_DGRAM, IPPROTO_UDP);
	free(addr);
	soutlen=sizeof(sout);
	if (getsockname(remotelisten, (SOCKADDR *)&sout, &soutlen)!=0) {
		perror("getsockname");
	}
	conn->sendResponsev5(0x00, (SOCKADDR *)&sout);

	memset(&sout, 0, sizeof(sout));
	((SOCKADDR *)&sout)->sa_family=addr->sa_family;
	soutlen=sizeof(sout);
	bind(remotetalk, (SOCKADDR *)&sout, soutlen);

	forwardUDPData(conn, remotelisten, remotetalk);
	CloseAndUnlink(remotetalk);
	CloseAndUnlink(remotelisten);

	return TRUE;
}

extern "C" {
BOOL command(void * conn)
{
	return cpp_command((Connection *)conn);
}
}


