
#include "../config.h"
#include <DSNetwork.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <unistd.h>

#define SUCCESS 3
#define PASSED  2
#define FAILED  0

void test_msg(char * msg)
{
	int i;
	printf("%s",msg);
	for (i=strlen(msg);i<65;i++)
		printf(".");
	fflush(NULL);
}

void test_res(int result)
{
	switch(result) {
		case SUCCESS:
			printf("Succeeded\n");
			break;
		case PASSED:
			printf("Passed\n");
			break;
		case FAILED:
			printf("FAILED\n");
			exit(EXIT_FAILURE);
			break;
	}
	fflush(NULL);
}

#pragma pack(1)

typedef struct socks4_req {
	unsigned char ver;
	unsigned char cmd;
	unsigned short dstport;
	unsigned long dstip;
	unsigned char end;
} socks4_req;

typedef struct socks4_resp {
	unsigned char ver;
	unsigned char cmd;
	unsigned short dstport;
	unsigned int dstip;
	unsigned char end;
} socks4_resp;

void test_fill_socks4_req(socks4_req * req)
{
	req->ver=0x4;
	req->cmd=0x01;
	req->dstport=0x00;
	req->dstip=0x00;
	req->end=0x00;
}

void socks4_test()
{
	DSNetwork net;
	DSNetwork temp;
	socks4_req req;
	socks4_resp resp;
	test_msg("Testing basic socks4");
	test_fill_socks4_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	req.dstport=htons(1080);
	req.dstip=htonl(0x7f000001);
	if (!net.sendData((char *)&req,sizeof(req))) test_res(FAILED);
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd==90) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	net.terminate();

	test_msg("Testing socks4 with user");
	test_fill_socks4_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	req.dstport=htons(1080);
	req.dstip=htonl(0x7f000001);
	if (!net.sendData((char *)&req,sizeof(req)-1)) test_res(FAILED);
	if (!net.sendData("username",9)) test_res(FAILED);
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd==90) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	net.terminate();


	test_msg("Testing socks4 with bad command");
	test_fill_socks4_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	req.cmd=0x65;
	req.dstport=htons(1080);
	req.dstip=htonl(0x7f000001);
	if (!net.sendData((char *)&req,sizeof(req))) test_res(FAILED);
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd==91) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	net.terminate();


	test_msg("Testing socks4 with bad address");
	test_fill_socks4_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	req.dstport=htons(1080);
	req.dstip=htonl(0x01010101);
	if (!net.sendData((char *)&req,sizeof(req))) test_res(FAILED);
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd==91) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	net.terminate();

	test_msg("Testing socks4 with bad port");
	test_fill_socks4_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	req.dstport=htons(65535);
	req.dstip=htonl(0x7f000001);
	if (!net.sendData((char *)&req,sizeof(req))) test_res(FAILED);
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd==91) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	net.terminate();

	test_msg("Testing basic socks4a");
	test_fill_socks4_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	req.dstport=htons(1080);
	req.dstip=htonl(0x00000001);
	if (!net.sendData((char *)&req,sizeof(req))) test_res(FAILED);
	if (!net.sendData("localhost",10)) test_res(FAILED);
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd==90) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	net.terminate();

	test_msg("Testing socks4a with user");
	test_fill_socks4_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	req.dstport=htons(1080);
	req.dstip=htonl(0x00000001);
	if (!net.sendData((char *)&req,sizeof(req)-1)) test_res(FAILED);
	if (!net.sendData("usrename",9)) test_res(FAILED);
	if (!net.sendData("localhost",10)) test_res(FAILED);
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd==90) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	net.terminate();

	test_msg("Testing bad socks4a hostname");
	test_fill_socks4_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	req.dstport=htons(1080);
	req.dstip=htonl(0x00000001);
	if (!net.sendData((char *)&req,sizeof(req))) test_res(FAILED);
	if (!net.sendData("badhostname",12)) test_res(FAILED);
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd==91) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	net.terminate();

	test_msg("Testing socks4 binding");
	test_fill_socks4_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	req.cmd=0x2;
	req.dstport=htons(1080);
	req.dstip=htonl(0x7f000001);
	if (!net.sendData((char *)&req,sizeof(req))) test_res(FAILED);
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd==90) {
		if (!temp.connectToHost("localhost",ntohs(resp.dstport))) test_res(FAILED);
		if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
		if (resp.cmd==90) {
			test_res(SUCCESS);
		} else {
			test_res(FAILED);
		}
	} else {
		test_res(FAILED);
	}
	net.terminate();

	test_msg("Testing socks4 binding with incorrect incoming addy");
	test_fill_socks4_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	req.cmd=0x2;
	req.dstport=htons(1080);
	req.dstip=htonl(0x01010101);
	if (!net.sendData((char *)&req,sizeof(req))) test_res(FAILED);
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd==90) {
		if (!temp.connectToHost("localhost",ntohs(resp.dstport))) test_res(FAILED);
		if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
		if (resp.cmd==91) {
			test_res(SUCCESS);
		} else {
			test_res(FAILED);
		}
	} else {
		test_res(FAILED);
	}
	net.terminate();



}

typedef struct socks5_ver {
	unsigned char ver;
	unsigned char nmeth;
	unsigned char meth;
} socks5_ver;

typedef struct socks5_ver_resp {
	unsigned char ver;
	unsigned char meth;
	unsigned char end;
} socks5_ver_resp;

typedef struct socks5_ip_req {
	unsigned char ver;
	unsigned char cmd;
	unsigned char rsv;
	unsigned char atyp;
	unsigned long dstip;
	unsigned short dstport;
} socks5_ip_req;

typedef struct socks5_name_req {
	unsigned char ver;
	unsigned char cmd;
	unsigned char rsv;
	unsigned char atyp;
	unsigned char nlen;
} socks5_name_req;

typedef struct socks5_resp {
	unsigned char ver;
	unsigned char cmd;
	unsigned char rsv;
	unsigned char atyp;
	unsigned int dstip;
	unsigned short dstport;
	unsigned char end;
} socks5_resp;

void test_fill_socks5_ver(socks5_ver * req)
{
	req->ver=0x5;
	req->nmeth=0x01;
	req->meth=0x00;
}

void test_fill_socks5_req(socks5_ip_req * req)
{
	req->ver=0x5;
	req->cmd=0x01;
	req->rsv=0x00;
	req->atyp=0x01;
	req->dstip=0x00;
	req->dstport=0x00;
}

void test_fill_socks5_name_req(socks5_name_req * req)
{
	req->ver=0x5;
	req->cmd=0x01;
	req->rsv=0x00;
	req->atyp=0x03;
	req->nlen=9;
}


void socks5_anon_ipv4_test()
{
	DSNetwork net;
	socks5_ver ver;
	socks5_ver_resp vr;
	socks5_ip_req req;
	socks5_resp resp;
	char szBuf[25];
	
	test_msg("Testing basic socks5");
	test_fill_socks5_ver(&ver);
	test_fill_socks5_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	if (!net.sendData((char *)&ver,sizeof(ver))) test_res(FAILED);
	if (!net.getSlab((char *)&vr,sizeof(vr)-1)) test_res(FAILED);
	if ((vr.meth!=ver.meth)||(vr.ver!=ver.ver)) test_res(FAILED);
	req.dstport=htons(1080);
	req.dstip=htonl(0x7f000001);
	if (!net.sendData((char *)&req,sizeof(req))) test_res(FAILED);
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd==0) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	net.terminate();

	test_msg("Testing socks5 with bad method");
	test_fill_socks5_ver(&ver);
	ver.meth=0x80;
	test_fill_socks5_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	if (!net.sendData((char *)&ver,sizeof(ver))) test_res(FAILED);
	if (!net.getSlab((char *)&vr,sizeof(vr)-1)) test_res(FAILED);
	if (vr.meth==0xFF) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	net.terminate();

	test_msg("Testing socks5 with bad addy");
	test_fill_socks5_ver(&ver);
	test_fill_socks5_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	if (!net.sendData((char *)&ver,sizeof(ver))) test_res(FAILED);
	if (!net.getSlab((char *)&vr,sizeof(vr)-1)) test_res(FAILED);
	if ((vr.meth!=ver.meth)||(vr.ver!=ver.ver)) test_res(FAILED);
	req.dstport=htons(1080);
	req.dstip=htonl(0x01010101);
	if (!net.sendData((char *)&req,sizeof(req))) test_res(FAILED);
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd!=0) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	net.terminate();

	test_msg("Testing socks5 with bad port");
	test_fill_socks5_ver(&ver);
	test_fill_socks5_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	if (!net.sendData((char *)&ver,sizeof(ver))) test_res(FAILED);
	if (!net.getSlab((char *)&vr,sizeof(vr)-1)) test_res(FAILED);
	if ((vr.meth!=ver.meth)||(vr.ver!=ver.ver)) test_res(FAILED);
	req.dstport=htons(65535);
	req.dstip=htonl(0x7f000001);
	if (!net.sendData((char *)&req,sizeof(req))) test_res(FAILED);
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd!=0) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	net.terminate();

	test_msg("Testing socks5 with user");
	test_fill_socks5_ver(&ver);
	ver.meth=0x02;
	test_fill_socks5_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	net.linger();
	if (!net.sendData((char *)&ver,sizeof(ver))) test_res(FAILED);
	if (!net.getSlab((char *)&vr,sizeof(vr)-1)) test_res(FAILED);
	if ((vr.meth!=ver.meth)||(vr.ver!=ver.ver)) test_res(FAILED);
	szBuf[0]=0x01;
	szBuf[1]=0x04;
	strcpy(&szBuf[2],"user");
	szBuf[6]=0x04;
	strcpy(&szBuf[7],"pass");
	if (!net.sendData((char *)szBuf,11)) test_res(FAILED);
	if (!net.getSlab((char *)szBuf,2)) test_res(FAILED);
	test_res(SUCCESS);
	net.terminate();
}

void socks5_anon_name_test()
{
	DSNetwork net;
	socks5_ver ver;
	socks5_ver_resp vr;
	socks5_name_req req;
	socks5_resp resp;
	unsigned short port;
	
	test_msg("Testing basic named socks5");
	test_fill_socks5_ver(&ver);
	test_fill_socks5_name_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	if (!net.sendData((char *)&ver,sizeof(ver))) test_res(FAILED);
	if (!net.getSlab((char *)&vr,sizeof(vr)-1)) test_res(FAILED);
	if ((vr.meth!=ver.meth)||(vr.ver!=ver.ver)) test_res(FAILED);
	if (!net.sendData((char *)&req,sizeof(req))) test_res(FAILED);
	if (!net.sendData((char *)"localhost",9)) test_res(FAILED);
	port=htons(1080);
	if (!net.sendData((char *)&port,2)) test_res(FAILED);
	
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd==0) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	net.terminate();

	test_msg("Testing with bad hostname");
	test_fill_socks5_ver(&ver);
	test_fill_socks5_name_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	if (!net.sendData((char *)&ver,sizeof(ver))) test_res(FAILED);
	if (!net.getSlab((char *)&vr,sizeof(vr)-1)) test_res(FAILED);
	if ((vr.meth!=ver.meth)||(vr.ver!=ver.ver)) test_res(FAILED);
	if (!net.sendData((char *)&req,sizeof(req))) test_res(FAILED);
	if (!net.sendData((char *)"xxxxxxxxx",9)) test_res(FAILED);
	port=htons(1080);
	if (!net.sendData((char *)&port,2)) test_res(FAILED);
	
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd!=0) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	net.terminate();
}


typedef struct socks5_udp_req {
	unsigned char rsv1;
	unsigned char rsv2;
	unsigned char frag;
	unsigned char atyp;
	unsigned long ipaddy;
	unsigned short port;
	char data[128];
} socks5_udp_req;

void socks5_udp_test()
{
	char szBuf1[512];
	SOCKET s,c;
	SOCKADDR sa;
	SOCKADDR_IN sin;
	SOCKADDR_IN srvsin;
	unsigned short endport;
	sl_t sl;
	socks5_ver ver;
	socks5_ver_resp vr;
	socks5_ip_req req;
	socks5_resp resp;
	socks5_udp_req udpreq;
	DSNetwork net;

	sl=sizeof(sa);
	s=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	c=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	memset(&sa,0,sizeof(sa));
	sa.sa_family=AF_INET;
	sl=sizeof(sa);
	bind(s, &sa, sl);
	memset(&sa,0,sizeof(sa));
	sa.sa_family=AF_INET;
	sl=sizeof(sa);
	bind(c, &sa, sl);
	sl=sizeof(sa);
	getsockname(s, &sa, &sl);
	endport=((SOCKADDR_IN *)&sa)->sin_port;

	test_msg("Testing socks5 UDP");
	test_fill_socks5_ver(&ver);
	test_fill_socks5_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	if (!net.sendData((char *)&ver,sizeof(ver))) test_res(FAILED);
	if (!net.getSlab((char *)&vr,sizeof(vr)-1)) test_res(FAILED);
	if ((vr.meth!=ver.meth)||(vr.ver!=ver.ver)) test_res(FAILED);
	req.dstport=((SOCKADDR_IN*)&sa)->sin_port;
	req.dstip=htonl(0x7f000001);
	req.cmd=0x03;
	if (!net.sendData((char *)&req,sizeof(req))) test_res(FAILED);
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd!=0) test_res(FAILED);
	memset(&udpreq,0,sizeof(udpreq));
	udpreq.atyp=0x01;
	udpreq.ipaddy=htonl(0x7f000001);
	udpreq.port=endport;
	strcpy(udpreq.data,"Hello");
	memset(&sin,0,sizeof(sin));
	sin.sin_family=AF_INET;
	sin.sin_port=resp.dstport;
	memcpy(&sin.sin_addr.s_addr,&resp.dstip,4);
	sl=sizeof(sin);
	sl=sendto(c, &udpreq, sizeof(udpreq), 0, (SOCKADDR *)&sin, sl);
	sl=sizeof(sin);
	recvfrom(s, &szBuf1, sizeof(szBuf1), 0, (SOCKADDR *)&srvsin, &sl);
	if (strcmp(szBuf1,"Hello")) test_res(FAILED);
	strcpy(szBuf1,"Greetings Earthlings");
	sl=sizeof(sin);
	sendto(s, szBuf1, strlen(szBuf1)+1, 0, (SOCKADDR *)&srvsin, sl);
	sl=sizeof(sin);
	recvfrom(c, &udpreq, sizeof(udpreq), 0, (SOCKADDR *)&sin, &sl);
	if (!strcmp(udpreq.data,"Greetings Earthlings")) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	
	net.terminate();

	test_msg("Testing socks5 UDP wo TCP addy");
	test_fill_socks5_ver(&ver);
	test_fill_socks5_req(&req);
	if (!net.connectToHost("localhost",1080)) test_res(FAILED);
	if (!net.sendData((char *)&ver,sizeof(ver))) test_res(FAILED);
	if (!net.getSlab((char *)&vr,sizeof(vr)-1)) test_res(FAILED);
	if ((vr.meth!=ver.meth)||(vr.ver!=ver.ver)) test_res(FAILED);
	req.dstport=0;
	req.dstip=0;
	req.cmd=0x03;
	if (!net.sendData((char *)&req,sizeof(req))) test_res(FAILED);
	if (!net.getSlab((char *)&resp,sizeof(resp)-1)) test_res(FAILED);
	if (resp.cmd!=0) test_res(FAILED);
	memset(&udpreq,0,sizeof(udpreq));
	udpreq.atyp=0x01;
	udpreq.ipaddy=htonl(0x7f000001);
	udpreq.port=endport;
	strcpy(udpreq.data,"Hello");
	memset(&sin,0,sizeof(sin));
	sin.sin_family=AF_INET;
	sin.sin_port=resp.dstport;
	memcpy(&sin.sin_addr.s_addr,&resp.dstip,4);
	sl=sizeof(sin);
	sl=sendto(c, &udpreq, sizeof(udpreq), 0, (SOCKADDR *)&sin, sl);
	sl=sizeof(sin);
	recvfrom(s, &szBuf1, sizeof(szBuf1), 0, (SOCKADDR *)&srvsin, &sl);
	if (strcmp(szBuf1,"Hello")) test_res(FAILED);
	strcpy(szBuf1,"Greetings Earthlings");
	sl=sizeof(sin);
	sendto(s, szBuf1, strlen(szBuf1)+1, 0, (SOCKADDR *)&srvsin, sl);
	sl=sizeof(sin);
	recvfrom(c, &udpreq, sizeof(udpreq), 0, (SOCKADDR *)&sin, &sl);
	if (!strcmp(udpreq.data,"Greetings Earthlings")) {
		test_res(SUCCESS);
	} else {
		test_res(FAILED);
	}
	
	net.terminate();

}



int main()
{
	socks4_test(); 
	printf("\n");
	socks5_anon_ipv4_test();
	socks5_anon_name_test();
	socks5_udp_test();
	return EXIT_SUCCESS;
}
